"""entity classes for vcsfile content types

:organization: Logilab
:copyright: 2007-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
import os.path as osp
from itertools import chain

from cubicweb import Unauthorized, ValidationError
from cubicweb.entities import AnyEntity, fetch_config

from cubes.vcsfile.hooks import repo_cache_dir

_MARKER = object()

def remove_authinfo(url):
    try:
        base, remaining = url.split('@')
    except ValueError:
        return url
    try:
        scheme, authinfo = base.split('://')
    except ValueError:
        return url
    return u'%s://***@%s' % (scheme, remaining)

class Repository(AnyEntity):
    """customized class for Repository entities"""
    __regid__ ='Repository'

    __permissions__ = ('write',)
    fetch_attrs, cw_fetch_order = fetch_config(['source_url', 'title', 'type'])
    rest_attr = 'eid' # see #XXX, using path cause pb w/ apache redirection

    @property
    def localcachepath(self):
        """return a path to the local repository cache (`.local_cache`)

        Does make sense only for mercurial repositories.
        The path itself should be enforced unless some late
        permissions problem arises (in which case this will fail).
        There is no warranty of an actual repository here.
        """
        if self.type != 'mercurial':
            raise TypeError('.localcachepath only makes sense on a mercurial repository')
        if self.local_cache is None:
            raise Exception('%s.local_cache does not exist yet' % self.eid)
        return osp.join(repo_cache_dir(self._cw.vreg.config), self.local_cache)

    def dc_title(self):
        if self.title:
            return self.title
        if self.source_url:
            return remove_authinfo(self.source_url)
        title = self._cw._('%(type)s repository #%(eid)s') % {
            'type': self.type, 'eid': self.eid}
        return title

    def printable_value(self, attr, *args, **kwargs):
        if attr == 'source_url':
            if args:
                value = args[0]
            elif 'value' in kwargs:
                value = kwargs['value']
            else:
                value = self.source_url
            if value:
                return remove_authinfo(value)
            return u''
        return super(Repository, self).printable_value(attr, *args, **kwargs)

    # navigation in versioned content #########################################

    def latest_known_revision(self):
        """return the changeset id of the latest known revision"""
        latestrev = self._cw.execute(
            'Any CS ORDERBY E DESC LIMIT 1 WHERE R eid %(r)s, REV from_repository R, '
            'REV changeset CS, REV eid E', {'r' : self.eid})[0][0]
        if latestrev is None:
            return -1
        return latestrev

    def heads_rset(self):
        """return a result set containing head revision (eg revisions without
        children)
        """
        return self._cw.execute(
            'Any REV WHERE R eid %(r)s, REV from_repository R, '
            'NOT CREV parent_revision REV, REV hidden FALSE', {'r' : self.eid})

    def branches(self):
        """return existing branches"""
        rql = 'DISTINCT Any B WHERE R eid %(r)s, REV from_repository R, REV branch B'
        branches = [b for b, in self._cw.execute(rql, {'r' : self.eid})]
        if not branches:
            branches = [self.default_branch()]
        return branches

    def branch_head(self, branch=_MARKER, public_only=False):
        """return latest revision of the given branch"""
        if branch is _MARKER:
            branch = self.default_branch()
        rql = (
            'Any MAX(REV) WHERE '
            '  REV from_repository R, R eid %%(r)s, REV branch %%(branch)s%s'
            % (', REV phase "public"' if public_only else ''))
        eid = self._cw.execute(rql, {'r': self.eid, 'branch': branch})[0][0]
        if eid is None:
            return None
        return self._cw.entity_from_eid(eid)

    # vcs write support ########################################################

    def default_branch(self):
        return u'default'

    def make_revision(self, msg, author=None, branch=None,
                      parentrev=None, added=(), deleted=()):
        """parameters:
        :msg: the check-in message (string, mandatory)
        :author: the check-in author (string)
        :branch: the branch in which the revision should be done (string)
        :parentrev: the parent of the revision
        :added: list of (filename, file-like) tuples
        :deleted: list of filenames to remove

        Filenames are relative to the repo root.

        Return the new revision entity.
        """
        if not self._cw.user.has_permission(u'write', self.eid):
            raise Unauthorized('You are not allowed to create new revisions')
        if author is None:
            author = self._cw.user.login
        assert parentrev, 'you must specify a parent revision'
        #assert parent
        for fpath in chain((fpath for fpath, _ in added), deleted):
            assert not (osp.isabs(fpath) or '../' in fpath), \
                'you must specify a path relative to the root of the repository, not %s' % fpath
        args = {'description' : msg, 'author': author, 'branch': branch,
                'parent': parentrev, 'added': added,
                'deleted': deleted}
        vcsrepo = self.cw_adapt_to('VCSRepo')
        csetid = vcsrepo.add_revision(**args)
        try:
            self.refresh(raise_on_error=True)
        except (Unauthorized, ValidationError):
            vcsrepo.strip_cset(csetid)
            raise
        rev = vcsrepo.entity_for_cset(csetid)
        with self._cw.security_enabled(write=False):
            rev.cw_set(created_by=self._cw.user)
        return rev

    def refresh(self, raise_on_error=False):
        self._cw.call_service('refresh_repository', eids=[self.eid], raise_on_error=raise_on_error)


class Revision(AnyEntity):
    """customized class for Revision entities"""
    __regid__ ='Revision'
    fetch_attrs, cw_fetch_order = fetch_config(['changeset', 'branch',
                                                'author', 'description', 'creation_date'],
                                               mainattr='eid',
                                               order='DESC')

    @property
    def shortdesc(self):
        return self.description.split('\n', 1)[0].strip()

    def dc_title(self):
        return u'#%(changeset)s %(desc)s' % {
            'changeset': self.changeset,
            'desc': self.shortdesc}

    def dc_long_title(self):
        try:
            return self._cw._('%(rev)s of repository %(repo)s: %(desc)s') % {
                'rev': self.dc_title(), 'repo': self.repository.dc_title(),
                'desc': self.shortdesc}
        except Unauthorized:
            return self._cw._('%(rev)s of a private repository') % {
                'rev': self.dc_title()}

    def sortvalue(self, rtype=None):
        if rtype is None:
            return self.eid
        return super(Revision, self).sortvalue(rtype)

    # navigation in versioned content #########################################

    @property
    def repository(self):
        return self.from_repository[0]

    @property
    def children_revisions(self):
        return self.reverse_parent_revision

    def export(self):
        '''return text/diff version of the revision'''
        if self.changeset is None:
            return None
        return self._cw.call_service('vcs.export-rev',
                                     repo=self.repository.eid,
                                     nodeid=self.changeset)




"""actions for entity types defined by the vcsfile package

:organization: Logilab
:copyright: 2007-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

try:
    from cubicweb import _
except ImportError:
    _ = unicode

from cubicweb.predicates import (one_line_rset, is_instance, rql_condition,
                                 score_entity, match_user_groups, anonymous_user)
from cubicweb.web.action import Action
from cubicweb.web.views.actions import CopyAction

class RepositoryRefreshAction(Action):
    __regid__ = 'refresh'
    __select__ = is_instance('Repository') & match_user_groups('managers')

    title = _('refresh repository')
    order = 130
    #category = 'moreactions'

    def url(self):
        entity = self.cw_rset.get_entity(0, 0)
        repoeid = entity.eid
        return self._cw.build_url('refresh_repository', eid=repoeid)


# deactivate copy action for entities coming from the svn source
CopyAction.__select__ = CopyAction.__select__ & ~is_instance('Revision')



def drop_entity_types_fast(*etypes, **kwargs):
    """drop entity types bypassing all hooks"""

    for etype in etypes:
        if etype not in schema:
            print '%s does not exist' % etype
            continue
        etype = schema[etype]

        # ignore attributes and inlined rels since they'll be dropped anyway
        srels = [x for x in etype.subject_relations() if x.eid and not (x.final or x.inlined)]

        orels = [x for x in etype.object_relations() if x.eid and not x.inlined]
        inlined_rels = [x for x in etype.object_relations() if x.eid and x.inlined]

        # eids to be deleted could be listed in some other entity tables through inlined relations
        for rtype in inlined_rels:
            for subjtype in rtype.subjects(etype):
                if subjtype in etypes:
                    continue
                print 'updating', subjtype.type
                rdef = rtype.rdef(subjtype, etype)
                if rdef.cardinality[0] == '1':
                    sql('DELETE FROM cw_%(stype)s WHERE cw_%(rtype)s IN (SELECT eid FROM entities WHERE type = %%s)'
                        % {'stype': subjtype.type, 'rtype': rtype.type},
                        (etype.type,))
                else:
                    sql('UPDATE cw_%(stype)s SET cw_%(rtype)s = NULL '
                        'WHERE cw_%(rtype)s IN (SELECT eid FROM entities WHERE type = %%s)' %
                        {'stype': subjtype.type, 'rtype': rtype.type},
                        (etype.type,))

        for rel in srels:
            print 'deleting from', rel.type
            if all(subj in etypes for subj in rel.subjects()) or all(obj in etypes for obj in rel.objects()):
                sql('DELETE FROM %s_relation' % rel.type)
            else:
                sql('DELETE FROM %s_relation WHERE eid_from IN (SELECT eid FROM entities WHERE type = %%s)' % rel.type, (etype.type,))
        for rel in orels:
            print 'deleting from', rel.type
            if all(subj in etypes for subj in rel.subjects()) or all(obj in etypes for obj in rel.objects()):
                sql('DELETE FROM %s_relation' % rel.type)
            else:
                sql('DELETE FROM %s_relation WHERE eid_to IN (SELECT eid FROM entities WHERE type = %%s)' % rel, (etype.type,))

        sql('DELETE FROM appears WHERE uid IN (SELECT eid FROM entities WHERE type = %s)', (etype.type,))
        sql('DELETE FROM cw_%s' % etype.type)
        sql('DELETE FROM entities WHERE type = %s', (etype.type,))

    for etype in etypes:
        drop_entity_type(etype, **kwargs)


if session.find('Repository', type=u'subversion'):
    print "subversion repositories are no longer supported, aborting"
    raise SystemExit(1)

# changed constraint
# must be done here (before the add/drop attributes)  otherwise
# the migration crashes for some reason I prefer not to know
sync_schema_props_perms('Revision')

drop_attribute('Revision', 'revision')

if 'path' in schema['Repository'].subject_relations():
    for vcsrepo in session.find('Repository').entities():
        if vcsrepo.source_url is not None:
            continue
        if vcsrepo.local_cache is not None:
            continue
        if vcsrepo.path is None:
            continue
        vcsrepo.cw_set(source_url=u'file://' + vcsrepo.path, local_cache=vcsrepo.path)

drop_attribute('Repository', 'path')
drop_attribute('Repository', 'subpath')
add_attribute('Revision', 'vcstags')

# changed description
sync_schema_props_perms('Repository')

# changed vocabulary
sync_schema_props_perms('type')

to_delete = [etype for etype in ('DeletedVersionContent', 'VersionContent', 'VersionedFile') if etype not in fsschema]
drop_entity_types_fast(*to_delete)
drop_relation_type('content_for')
drop_relation_type('from_revision')
drop_relation_type('at_revision')
drop_relation_type('vc_copy')
drop_relation_type('vc_rename')



"""cubicweb-ctl plugin providing the vcscheck command

:organization: Logilab
:copyright: 2013-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from __future__ import print_function

__docformat__ = "restructuredtext en"

from cubicweb.cwctl import CWCTL
from cubicweb.toolsutils import Command

class VCSRefreshCommand(Command):
    """ Refresh repository

    <instance id>
      identifier of the instance
    """
    name = 'vcsrefresh'
    min_args = max_args = 1
    arguments = '<instance id>'

    def run(self, args):
        """run the command with its specific arguments"""
        from cubicweb.server.serverconfig import ServerConfiguration
        from cubicweb.server.serverctl import repo_cnx
        appid = args.pop(0)
        config = ServerConfiguration.config_for(appid)
        config.__class__.cube_appobject_path = set(('hooks', 'entities'))
        config.__class__.cubicweb_appobject_path = set(('hooks', 'entities'))
        repo, cnx = repo_cnx(config)

        with cnx:
            rset = cnx.execute('Any X,S where X is Repository, X source_url S')
            for hgrepo, source_url in rset:
                print('refreshing', source_url)
                cnx.call_service('refresh_repository', eids=(hgrepo,))

CWCTL.register(VCSRefreshCommand)



"""primary views for entity types defined by the vcsfile package

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

try:
    from cubicweb import _
except ImportError:
    _ = unicode

from logilab.mtconverter import TransformError, xml_escape, TransformData

from cubicweb import Unauthorized, tags
from cubicweb.mttransforms import ENGINE
from cubicweb.predicates import is_instance, score_entity, match_kwargs
from cubicweb.view import EntityView
from cubicweb.uilib import soup2xhtml, domid
from cubicweb.web.views import (ibreadcrumbs, idownloadable, tableview,
                                baseviews, primary, tabs, navigation, uicfg)

# primary view tweaks
_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl
_abaa = uicfg.actionbox_appearsin_addmenu

# internal purpose relation
_pvs.tag_attribute(('Repository', 'local_cache'), 'hidden')
_pvs.tag_attribute(('Repository', 'source_url'), 'hidden') # custom view
_pvs.tag_object_of(('*', 'from_repository', 'Repository'), 'hidden')

_pvs.tag_subject_of(('Revision', 'parent_revision', 'Revision'), 'attributes')
_pvs.tag_object_of(('Revision', 'parent_revision', 'Revision'), 'attributes')
_pvs.tag_subject_of(('Revision', 'from_repository', '*'), 'hidden') # in breadcrumb

_pvdc.tag_subject_of(('*', 'from_repository', '*'), {'vid': 'incontext'})

_pvdc.tag_attribute(('Repository', 'source_url'), {'vid': 'urlattr'})

_pvdc.tag_attribute(('Revision', 'description'),
                    {'showlabel': False, 'vid': 'vcsfile.rev.description'})

# we don't want automatic addrelated action for the following relations...
for rtype in ('from_repository', 'parent_revision'):
    _abaa.tag_object_of(('*', rtype, '*'), False)


def render_entity_summary(self, entity):
    if not entity.is_head(entity.rev.branch):
        msg = self._cw._('this file has newer revisions')
        self.w(tags.div(msg, klass='warning'))
    if entity.description:
        self.field(self._cw._('commit message:'), xml_escape(entity.description),
                   tr=False, table=False)


class RepositoryPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('Repository')

    tabs = [_('repositoryinfo_tab'), _('repositoryhistory_tab')]
    default_tab = 'repositoryinfo_tab'


class RepositoryInfoTab(tabs.PrimaryTab):
    __regid__ = 'repositoryinfo_tab'
    __select__ = is_instance('Repository')

    def render_entity_attributes(self, entity):
        super(RepositoryInfoTab, self).render_entity_attributes(entity)
        if entity.source_url:
            entity.view('vcsfile.repository.checkout', w=self.w)
        rset = self._cw.execute('Any REV, REVB ORDERBY REV DESC '
                                'WHERE REV branch REVB, REV from_repository R, '
                                'R eid %(r)s, NOT X parent_revision REV, REV hidden FALSE',
                                {'r': entity.eid})
        if rset:
            self.w('<h3>%s</h3>' % self._cw._('heads'))
            self.wview('table', rset)


class RepositoryHistoryTab(EntityView):
    __regid__ = 'repositoryhistory_tab'
    __select__ = is_instance('Repository')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rset = self._cw.execute(
            'Any R,RB,RA,RD,RCD, RCS ORDERBY R DESC WHERE '
            'R branch RB, R author RA, R description RD, R creation_date RCD,'
            'R changeset RCS, R from_repository X, X eid %(x)s',
            {'x': entity.eid})
        self.wview('vcsfile.table.revisions', rset, 'noresult')

class RevisionsTable(tableview.RsetTableView):
    __regid__ = 'vcsfile.table.revisions'
    __select__ = is_instance('Revision')
    displaycols = range(5)
    layout_args = {'display_filter': 'top'}


class RepositoryCheckOutInstructionsView(EntityView):
    __regid__ = 'vcsfile.repository.checkout'
    __select__ = is_instance('Repository') & score_entity(lambda x: x.source_url)

    def entity_call(self, entity):
        w = self.w
        _ = self._cw._
        url_scheme = entity.source_url.split(':')[0]
        if url_scheme.startswith('http'):
            w(u'<h3>%s</h3>' % _('Browse source'))
            w(u'<p>%s</p>' % _('You can browse the source code by following <a href="%s">this link</a>.')
              % xml_escape(entity.source_url))
        w(u'<h3>%s</h3>' % _('Command-Line Access'))
        if entity.has_anonymous_access:
            msg = _('Non-members may check out a read-only working copy anonymously over %s.') % xml_escape(url_scheme.upper())
        else:
            msg = _('Members only may check out working copy over %s.') % xml_escape(url_scheme.upper())
        w(u'<p>%s</p>' % msg)
        w(u'<p>%s</p>' % _('Use this command to check out the latest project source code:'))
        w(u'<pre>')
        if entity.type == 'mercurial':
            w(u'hg clone %s' % xml_escape(entity.source_url))
        w(u'</pre>')


class RevisionShortView(EntityView):
    __regid__ = 'shorttext'
    __select__ = is_instance('Revision')
    content_type = 'text/plain'

    def cell_call(self, row, col):
        rev = self.cw_rset.get_entity(row, col)
        self.w(u'#%s' % rev.changeset)


class RevisionPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Revision')

    def render_entity_title(self, entity):
        self.w(u'<h1>%s</h1>' % xml_escape(entity.shortdesc))

    def render_entity_relations(self, rev):
        repo = rev.repository.cw_adapt_to('VCSRepo')
        versioned = repo.status(rev.changeset)
        if versioned:
            tpl = u'<a href="#%s">%s</a>\n'
            headers = (self._cw._('files modified by this revision'),)
            pyvalue = [[tpl % (domid(obj), xml_escape(obj))] for obj in versioned]
            self.wview('pyvaltable', pyvalue=pyvalue, headers=headers)
        else:
            self.w(u'<p>%s</p>' % self._cw._('this revision hasn\'t modified any files'))
        patch = rev.export()
        if patch is not None:
            self.w(u'<div class="content">')
            transformer = rev._cw_mtc_transform
            html = transformer(patch, 'text/x-diff', 'text/annotated-html', 'utf8')
            self.w(html)
            self.w(u'</div>')


class RevisionDescriptionView(EntityView):
    __regid__ = 'vcsfile.rev.description'
    __select__ = EntityView.__select__ & is_instance('Revision') & match_kwargs('rtype')

    def entity_call(self, entity, rtype, **kwargs):
        if '\n' not in entity.description:
            return
        longdesc = entity.description.split('\n', 1)[1].lstrip('\n')
        trdata = TransformData(longdesc, 'text/rest', 'utf8', appobject=self)
        data = ENGINE.convert(trdata, 'text/html').decode()
        data = soup2xhtml(data, self._cw.encoding)
        if '<div class="system-message">' not in data:
            # this is what cubicweb/ext/html4zope.py injects when there are
            # parsing errors. Couldn't find a way to get an exception instead...
            self.w(data)
        else:
            self.w(u'<p>\n')
            self.w(xml_escape(longdesc).replace('\n', '<br>'))
            self.w(u'</p>\n')


# navigation: breadcrumbs / prevnext adapters ##################################

class RepoContentIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('Revision')

    def parent_entity(self):
        try:
            return self.entity.repository
        except Unauthorized:
            return None


class RevisionBreadCrumbTextView(ibreadcrumbs.BreadCrumbTextView):
    __select__ = is_instance('Revision')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        self.w(entity.changeset)


class RevisionIPrevNextAdapter(navigation.IPrevNextAdapter):
    __select__ = is_instance('Revision')

    def previous_entity(self):
        # may have multiple parents, give priority to the one in the same
        # branch.
        parent = None
        for parent in self.entity.parent_revision:
            if parent.branch == self.entity.branch:
                return parent
        return parent

    def next_entity(self):
        # may have multiple children, give priority to the one in the same
        # branch.
        child = None
        for child in self.entity.reverse_parent_revision:
            if child.branch == self.entity.branch:
                return child
        return child



# repository specific stuff ####################################################

try:
    from cubicweb import server
except ImportError: # no server installation
    pass
else:
    options = (
        ('repository-import',
         {'type' : 'yn',
          'default': True,
          'help': 'Is the instance responsible to automatically import new '
                   'revisions from repositories? '
                   'You should say yes unless you don\'t want this behaviour '
                   'or if you use a multiple repositories setup, in which '
                   'case you should say yes on one repository, no on others.',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('check-revision-interval',
         {'type' : 'int',
          'default': 5*60,
          'help': 'interval between checking of new revisions in repositories \
(default to 5 minutes).',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('check-revision-commit-every',
         {'type' : 'int',
          'default': 1000,
          'help': 'after how much new imported revisions the transaction \
should be commited (default to 1000).',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('local-repo-cache-root',
         {'type':'string',
          'default': 'repo_cache', # XXX /var/lib/cubicweb/<instance>
          'help':'Local repository cache location (if not absolute, will be '
          'relative to instance data home directory).',
          'level': 2,
          'group': 'vcsfile'
          }
         ),
        ('cache-external-repositories',
         {'type':'yn',
          'default': False,
          'help':'Should repositories from external source have a local cache?'
          'This is usually not necessary, beside cases where for instance a '
          'narval bot on the same host as the instance could benefit from them '
          'to run apycot tests.',
          'level': 2,
          'group': 'vcsfile'
          }
         ),
        ('hgrc-path',
         {'type': 'string',
          'default': None,
          'help': 'a list of files or directories to search for mercurial configuration',
          'level': 2,
          'group': 'vcsfile',
          }
         ),
        )

from docutils import nodes, statemachine
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.misc import Include

from cubes.vcsfile.entities import Repository
from cubes.vcsfile.docparser import Diff2HTMLTransform
from cubicweb.mttransforms import ENGINE

COMPONENT_CONTEXT = 'navcontentbottom'

if Diff2HTMLTransform.name not in ENGINE.transforms:
    ENGINE.add_transform(Diff2HTMLTransform())


class VCSInclude(Include):
    option_spec = Include.option_spec.copy()
    option_spec['revision'] = str
    option_spec['repository'] = int

    def run(self):
        # This is a partial copy of Include.run method, but
        # docutils...
        entity = self.state.document.settings.context
        repoeid = self.options.get('repository')
        if repoeid is None:
            try:
                repo = entity.repository
            except AttributeError:
                raise self.severe("vcsinclude directive requires the 'repository' option"
                                  "when used in the context of a %s" %
                                  entity.__class__.__name__)
        else:
            repo = entity._cw.entity_from_eid(repoeid)
        if not isinstance(repo, Repository):
            raise self.severe('\'repository\' option value is not the eid of a '
                              'Repository but a %s.' % repo.__class__.__name__)

        path = self.arguments[0]

        rev = self.options.get('revision')
        if rev is None:
            rev = 'tip'
        vcsrepo = repo.cw_adapt_to('VCSRepo')
        try:
            content = vcsrepo.cat(rev, path)
        except Exception:
            try:
                content = vcsrepo.cat('tip', path)
            except Exception as e:
                raise self.severe(e)
            self.warning('Could not get included file at revision "%s", '
                         'using latest revision instead' % rev)

        if content is None:
            return []

        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        encoding = self.options.get('encoding', 'utf-8')
        tab_width = self.options.get(
            'tab-width', self.state.document.settings.tab_width)
        startline = self.options.get('start-line', None)
        endline = self.options.get('end-line', None)

        rawtext = content.decode(encoding)

        if startline or (endline is not None):
            lines = rawtext.splitlines(True)
            rawtext = ''.join(lines[startline:endline])

        # start-after/end-before: no restrictions on newlines in match-text,
        # and no restrictions on matching inside lines vs. line boundaries
        after_text = self.options.get('start-after', None)
        if after_text:
            # skip content in rawtext before *and incl.* a matching text
            after_index = rawtext.find(after_text)
            if after_index < 0:
                raise self.severe('Problem with "start-after" option of "%s" '
                                  'directive:\nText not found.' % self.name)
            rawtext = rawtext[after_index + len(after_text):]
        before_text = self.options.get('end-before', None)
        if before_text:
            # skip content in rawtext after *and incl.* a matching text
            before_index = rawtext.find(before_text)
            if before_index < 0:
                raise self.severe('Problem with "end-before" option of "%s" '
                                  'directive:\nText not found.' % self.name)
            rawtext = rawtext[:before_index]
        if 'literal' in self.options:
            # Convert tabs to spaces, if `tab_width` is positive.
            if tab_width >= 0:
                text = rawtext.expandtabs(tab_width)
            else:
                text = rawtext
            literal_block = nodes.literal_block(rawtext, text, source=path)
            literal_block.line = 1
            return [literal_block]
        else:
            include_lines = statemachine.string2lines(
                rawtext, tab_width, convert_whitespace=1)
            self.state_machine.insert_input(include_lines, path)
            return []


directives.register_directive('vcsinclude', VCSInclude)




from hglib.error import CommandError

from cubicweb.server import Service


class RevisionExportService(Service):
    """return the patch version of a revision
    """

    __regid__ = 'vcs.export-rev'

    def call(self, repo, nodeid):
        repo_ent = self._cw.entity_from_eid(repo)
        if repo_ent.type != 'mercurial':
            # no svn support yet (use a selector when we have multiple versions)
            return None
        hdrepo = repo_ent.cw_adapt_to('VCSRepo')
        try:
            with hdrepo.hgrepo() as hgrepo:
                if nodeid in hgrepo:
                    return hgrepo.export(revs=nodeid, git=True)
        except CommandError as e:
            return None


class RevisionDiffService(Service):
    """ Return the diff between two revisions."""

    __regid__ = 'vcsfile.export-rev-diff'

    def call(self, repo_eid, rev1, rev2):
        repo = self._cw.entity_from_eid(repo_eid)
        adapter = repo.cw_adapt_to('VCSRepo')
        try:
            with adapter.hgrepo() as hgrepo:
                return hgrepo.diff(revs=[rev1, rev2], git=True, unified=5,
                                   showfunction=True)
        except CommandError:
            return None



if 'ancestor_revision' not in fsschema:
    # some apps might want to keep it
    sql('DELETE FROM ancestor_revision_relation') # quick delete
    drop_relation_definition('Revision', 'ancestor_revision', 'Revision')

for rdef in ('Revision.description', 'Revision.author',
             'VersionedFile.directory', 'VersionedFile.name'):
    rdef = rdef.split('.')
    if len(rdef) < 3:
        rdef.append('String')
    sync_schema_props_perms(tuple(rdef), syncperms=False)



"""entity types used to represent a version control system (vcs) content

:organization: Logilab
:copyright: 2007-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

try:
    from cubicweb import _
except ImportError:
    _ = unicode

from yams.buildobjs import (DEFAULT_ATTRPERMS,
                            EntityType, RelationDefinition,
                            String, Boolean)
from cubicweb.schema import RRQLExpression

PRIVATE_ATTRPERMS = DEFAULT_ATTRPERMS.copy()
PRIVATE_ATTRPERMS['read'] = ('managers',)
PRIVATE_ATTRPERMS['update'] = ('managers',)

RO_ATTRPERMS = DEFAULT_ATTRPERMS.copy()
RO_ATTRPERMS['update'] = ()
RO_RELPERMS = DEFAULT_ATTRPERMS.copy()
RO_RELPERMS['add'] = (RRQLExpression('X from_repository R, U in_group G, '
                                     'R require_permission P, P require_group G'))


class Repository(EntityType):
    """a local VCS repository which will be used as versionned content
    entity source
    """
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers',),
        'update': ('managers',),
        'delete': ('managers',),
        }

    title = String(description=_('optional title. If the repository has a public url, '
                                 'it may be a good idea to put it as title.'),
                   maxsize=256)
    type = String(required=True, vocabulary=('mercurial',),
                  description=_('repository\'s type'))
    source_url = String(description=_('source url (must be of the form protocol://path/to/stuff)'),
                        unique=True)
    local_cache = String(unique=True,
                         description=_('relative path to the local cache of the repository'),
                         __permissions__=PRIVATE_ATTRPERMS)
    encoding = String(default='utf-8', maxsize=20, required=True,
                      description=_('encoding used for the repository (e.g.'
                                    ' for file path and check-in comments)'))
    has_anonymous_access = Boolean(default=True,
                                   description=_('Tell if user needs an account to checkout the repository.'))


class Revision(EntityType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (),
        'update': (),
        'delete': (),
        }
    __unique_together__ = [('from_repository', 'changeset')]
    # NOTE: creation_date will be set at revision date

    description = String(description=_('comment for this revision.'))

    author = String(indexed=True, maxsize=200,
                    description=_("author of this revision."))

    changeset = String(indexed=True, maxsize=100,
                       description=_('change set identifier, when used by the '
                                     'underlying version control system.'),
                       fulltextindexed=True,
                       __permissions__=RO_ATTRPERMS)

    branch = String(indexed=True, maxsize=255,
                    description=_("branch of this revision."),
                   __permissions__=RO_ATTRPERMS)

    phase = String(description=_('phase of the changeset'), indexed=True,
                   default='draft', vocabulary=[_('secret'), _('draft'), _('public')])

    vcstags = String(indexed=True, maxsize=255,
                     description=_('Tags on this revision (comma separated values)'),
                     __permissions__=RO_ATTRPERMS)

    hidden = Boolean(required=True, indexed=True)


class from_repository(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (),
        'delete': (),
        }
    inlined = True
    subject = 'Revision'
    object = 'Repository'
    cardinality = '1*'
    composite = 'object'


class parent_revision(RelationDefinition): # XXX ordered
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (),
        'delete': (),
        }
    subject = 'Revision'
    object = 'Revision'

class obsoletes(RelationDefinition):
    __permissions__ = {'read': ('managers', 'users', 'guests'),
                       'add': (),
                       'delete': ()}
    subject = 'Revision'
    object = 'Revision'

class granted_permission(RelationDefinition):
    subject = 'Repository'
    object = 'CWPermission'

class require_permission(RelationDefinition):
    subject = 'Repository'
    object = 'CWPermission'



import sys
import os, os.path as osp
import shutil

import hglib

from logilab.common.decorators import classproperty
from logilab.common.shellutils import unzip

from cubicweb.devtools.testlib import CubicWebTC

from cubes.vcsfile import bridge

DATA = osp.abspath(osp.join(osp.dirname(__file__), 'data'))

def setup_repos(*reponames):
    for repo in reponames:
        if isinstance(repo, unicode):
            repo = repo.encode('utf-8')
        if osp.exists(repo):
            shutil.rmtree(repo)
        if osp.exists('%s.zip' % repo):
            unzip('%s.zip' % repo, osp.dirname(repo))
        else: # no zip file, let's init a new repo
            hglib.init(repo)

def init_vcsrepo(repo, commitevery=1):
    with repo.internal_cnx() as cnx:
        bridge.import_content(cnx, commitevery=commitevery,
                              raise_on_error=True)

class VCSRepositoryTC(CubicWebTC):
    """base class for test which should import a repository during setup"""
    _repo_path = None
    repo_type = u'mercurial'
    repo_encoding = u'utf-8'
    repo_title = None

    commitevery = 3

    @classproperty
    def test_db_id(cls):
        if cls._repo_path is None:
            return None
        return cls._repo_path

    @classproperty
    def repo_path(cls):
        assert cls._repo_path, 'you must set repository through _repo_path first'
        return osp.join(cls.datadir, cls._repo_path)

    @classmethod
    def _create_repo(cls, cnx):
        return cnx.create_entity(
            'Repository', type=cls.repo_type, source_url='file://'+cls.repo_path,
            encoding=cls.repo_encoding,
            title=cls.repo_title)

    @classmethod
    def grant_write_perm(cls, session, group, vcsrepoeid=None):
        if vcsrepoeid is None:
            vcsrepoeid = cls.vcsrepoeid
        cls.grant_permission(session, vcsrepoeid, group, u'write',
                             u'repo x write perm')

    @classmethod
    def pre_setup_database(cls, cnx, config):
        setup_repos(cls.repo_path)
        # don't use cls.vcsrepo in regular test, only in pre_setup_database
        cls.vcsrepoeid = cls._create_repo(cnx).eid
        cnx.commit()

    def setUp(self):
        setup_repos(self.repo_path)
        super(VCSRepositoryTC, self).setUp()
        self.refresh()
        with self.admin_access.repo_cnx() as cnx:
            vcsrepo = cnx.find('Repository').one()
            self.vcsrepoeid = vcsrepo.eid
            self.vcsrepopath = vcsrepo.localcachepath

    def tearDown(self):
        super(VCSRepositoryTC, self).tearDown()
        shutil.rmtree(self.vcsrepopath, ignore_errors=True)

    def refresh(self):
        with self.repo.internal_cnx() as cnx:
            bridge.import_content(cnx, commitevery=1, raise_on_error=True)
        with hglib.open(self.repo_path) as repo:
            return dict((int(rev.rev), rev.node[:12].decode('ascii'))
                        for rev in repo.log(hidden=True))


class HGRCMixin(object):
    _old_files = ()
    hgconfig = {'ui': {'username': 'babar',},
                'extensions': {'rebase': '',
                               'evolve': '',},
                'phases': {'publish': 'False',},
                }
    def setUp(self):
        hgrcpath = osp.join(self.datadir, 'obsolete.hgrc')
        self._oldhgrc = os.environ.get('HGRCPATH')

        with open(hgrcpath, 'wb') as hgrc:
            for section, values in self.hgconfig.items():
                print >> hgrc, '[%s]'%section.strip()
                for attr, value in values.items():
                    print >> hgrc, '%s=%s'%(attr, value)
        os.environ['HGRCPATH'] = hgrcpath
        self._old_files = [hgrcpath]
        super(HGRCMixin, self).setUp()

    def tearDown(self):
        super(HGRCMixin, self).tearDown()
        if self._oldhgrc is None:
            os.environ.pop('HGRCPATH')
        else:
            os.environ['HGRCPATH'] = self._oldhgrc
        for path in self._old_files:
            os.unlink(path)

class VCSFileTC(HGRCMixin, VCSRepositoryTC):
    _repo_path = u'testrepo'



# copyright 2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import re

from logilab.mtconverter import xml_escape
from logilab.mtconverter.transform import Transform

from cubicweb.uilib import domid


class Diff2HTMLTransform(Transform):
    """note: this transform use CSS classes which can be obtained using:
    formatter.get_style_defs()
    """
    inputs = 'text/x-diff',
    output = 'text/annotated-html'
    name = 'Diff2HTMLTransform'

    def __init__(self, ipcb=None):
        # insert point callback
        self._ipcb = ipcb

    def _convert(self, trdata):
        cw = trdata.appobject._cw
        trdata.__counter = 0
        cw.add_css('cubes.vcsfile.css')
        output = []
        w = output.append
        cbs = {'+++': self.new_file,
               '---': self.old_file,
               'diff': self.start_file,
               '@@': self.start_chunk,
               }
        w(u'<div class="text/x-diff">\n')
        w(u'<div class="diffComment">\n')
        self._done_cb = self.stop_comment
        self._continue_cb = self.comment_line
        lines = trdata.decode().split('\n')
        if lines[-1] == '':
            del lines[-1]
        for line in lines:
            try:
                cb = cbs[line.split(' ', 1)[0]]
            except KeyError:
                if self._continue_cb is not None:
                    self._continue_cb(trdata, w, line)
            else:
                if self._done_cb is not None:
                    self._done_cb(trdata, w)
                    self._done_cb = None
                self._continue_cb = None
                cb(trdata, w, line)
        if self._done_cb is not None:
            self._done_cb(trdata, w)
        w(u'</div>\n')
        return ''.join(output)

    def diff_line(self, trdata, w, line):
        cssclass = {'-': 'diffOldLine',
                    '+': 'diffNewLine',
                    ' ': 'diffLine',
                    '\\': 'diffOldLine'}[line[0]]
        trdata.__counter += 1
        w(u'<span class="lineCount">%s</span> ' % trdata.__counter)
        if line.strip():
            w(u'<span class="%s">%s</span>\n' % (cssclass, xml_escape(line)))
        else:
            w('\n')

    def comment_line(self, trdata, w, line):
        w(xml_escape(line) + '<br/>')

    def new_file(self, trdata, w, line):
        #w(u'<span class="diffFile">%s</span>\n' % xml_escape(line))
        pass

    def old_file(self, trdata, w, line):
        pass

    _start_file_rgx = re.compile(r'.* a/(.*?) b/(.*)')

    def start_file(self, trdata, w, line):
        mobj = self._start_file_rgx.match(line)
        a, b = mobj.groups() if mobj is not None else ('', '')
        if a:
            a = u'<a name="%s"></a>\n' % domid(a)
        if b:
            b = u'<a name="%s"></a>\n' % domid(b)
        w(u'<div class="diffCmd">%s%s%s</div>\n' % (a, b, xml_escape(line)))

    def start_chunk(self, trdata, w, line):
        w(u'<div class="diffChunk">\n')
        w(u'<span class="diffChunkTitle">%s</span>\n' % xml_escape(line))
        w(u'<pre class="diff">\n')
        self._done_cb = self.stop_chunk
        self._continue_cb = self.diff_line

    def inc_counter(self, trdata):
        try:
            trdata.insert_points_count += 1
            ipid = trdata.insert_points_count
        except AttributeError:
            ipid = trdata.insert_points_count = 1
        return ipid

    def stop_chunk(self, trdata, w):
        w(u'</pre>\n')
        ipid = self.inc_counter(trdata)
        if self._ipcb is not None:
            self._ipcb(ipid, trdata, w)
        w(u'</div>\n')

    def stop_comment(self, trdata, w):
        ipid = self.inc_counter(trdata)
        if self._ipcb is not None:
            self._ipcb(ipid, trdata, w)
        w(u'</div>\n')



# -*- coding: utf-8 -*-
"""entity classes for vcsfile content types

:organization: Logilab
:copyright: 2007-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
import os, os.path as osp
import tempfile
import shutil
import errno
from contextlib import contextmanager
from itertools import izip_longest

import hglib
from hglib import context
from hglib.error import CommandError
from hglib.util import cmdbuilder

from yams import ValidationError
from cubicweb import QueryError
from cubicweb.predicates import is_instance, score_entity
from cubicweb.view import EntityAdapter

from cubes.vcsfile import bridge
from cubes.vcsfile.hooks import repo_cache_dir


class HGRepo(EntityAdapter):
    __regid__ = 'VCSRepo'
    __select__ = (is_instance('Repository') &
                  score_entity(lambda repo: repo.type == u'mercurial'))
    configs = ['extensions.hgext.strip=', 'extensions.hgext.evolve=']  # enable strip command
    path = None

    def entity_for_cset(self, ctx_hex):
        rset = self._cw.execute('Any X WHERE X from_repository R, R eid %(repo)s, X changeset %(cs)s',
                                {'repo': self.entity.eid, 'cs': unicode(ctx_hex)})
        if rset:
            return rset.one()
        return None

    def full_path(self, path):
        return osp.join(self.path, path)

    def __contains__(self, path):
        with self.hgrepo() as hgclient:
            return path in hgclient.manifest(all=True)

    def _create_repo(self):
        if osp.exists(self.path):
            return
        hglib.init(self.path)

    def __init__(self, *args, **kwargs):
        super(HGRepo, self).__init__(*args, **kwargs)
        cnx = self._cw
        try:
            security_enabled = cnx.security_enabled
        except AttributeError:
            # XXX Request objects don't have that method in cw 3.19
            security_enabled = cnx.cnx.security_enabled

        cacheroot = repo_cache_dir(self.entity._cw.vreg.config)
        # deactivate read security first, needed to get access to local_cache
        with security_enabled(read=False):
            if cacheroot is not None and self.entity.local_cache is not None:
                self.path = osp.join(cacheroot, self.entity.local_cache)

        if not self.path:
            raise ValidationError(self.entity.eid,
                    {'source_url': cnx._('no local access to repository')})

        if not osp.exists(self.path):
            self._create_repo() # ensure local cache exists

    @contextmanager
    def hgrepo(self):
        client = hglib.client.hgclient(self.path, encoding=self.entity.encoding,
                                       configs=self.configs, connect=False)
        client.hidden = True
        with client:
            yield client

    def manifest(self, rev):
        with self.hgrepo() as hgclient:
            return list(hgclient.manifest(rev=rev))

    def file_deleted(self, path, rev=None):
        with self.hgrepo() as repo:
            if rev is None:
                log = repo.log(files=[self.full_path(path)], removed=True, limit=1)
                if not log:
                    raise Exception('this file does not even exist')
                rev = log[0][1][:12]
            for file in repo.manifest(rev=rev):
                if file[4] == path:
                    return False
            return True

    def next_versions(self, rev, path, **args):
        with self.hgrepo() as repo:
            return repo.log(files=[self.full_path(path)], revrange='%s::' % rev.changeset,
                            **args)

    def previous_versions(self, rev, path, **args):
        with self.hgrepo() as repo:
            return repo.log(files=[self.full_path(path)], revrange='::%s' % rev.changeset,
                            **args)

    def root(self):
        with self.hgrepo() as repo:
            return repo.root()

    def head(self, path=None, **args):
        if path is not None:
            assert 'files' not in args
            args['files'] = [self.full_path(path)]
        with self.hgrepo() as repo:
            try:
                logs = repo.log(limit=1, removed=True, **args)
            except CommandError as exc:
                if 'unknown revision' in exc.err:
                    return None
                raise
        if logs:
            return logs[0]
        return None

    def log(self, path=None, **args):
        args.setdefault('removed', True)
        with self.hgrepo() as repo:
            if path is not None:
                args['files'] = [self.full_path(path)]
            return repo.log(**args)

    def log_rset(self, path=None, **args):
        """Returns a ResultSet of the Revisions
        """
        args.setdefault('removed', True)
        log = self.log(path, **args)
        # XXX what a funny joke
        # we need to split the big IN into several queries to
        # prevent a "Too many SQL variables" error
        rsets = []
        # itertools grouper pattern
        args = [iter(log)] * 50
        for partiallog in izip_longest(fillvalue=None, *args):
            rql = ('Revision R ORDERBY R DESC WHERE '
                   'R from_repository REPO, REPO eid %%(x)s, R changeset IN (%s)'
                   % ','.join('"%s"' % rev.node[:12].decode('ascii')
                              for rev in partiallog if rev is not None))
            rsets.append(self._cw.execute(rql,
                                          {'x': self.entity.eid}))
        if rsets:
            return reduce(lambda x, y: x+y, rsets)
        return self._cw.empty_rset()

    def status(self, cset):
        with self.hgrepo() as repo:
            return repo[cset].files()

    def cat(self, rev, path):
        with self.hgrepo() as repo:
            return repo.cat([self.full_path(path)], rev=rev)

    def strip_cset(self, cset):
        with self.hgrepo() as repo:
            args = cmdbuilder('strip', cset)
            repo.rawcommand(args)

    def add_revision(self, description, author, branch, parent, added, deleted):
        tmpdir = tempfile.mkdtemp()
        hglib.clone(self.path, tmpdir, updaterev=parent)
        try:
            with hglib.open(tmpdir) as client:
                if branch is not None:
                    client.branch(branch, force=True)
                for (file, content) in added:
                    path = osp.realpath(osp.join(tmpdir, file))
                    assert path.startswith(osp.realpath(tmpdir))
                    try:
                        os.makedirs(osp.dirname(path))
                    except OSError as exc:
                        if exc.errno != errno.EEXIST:
                            raise
                    with open(path, 'wb') as f:
                        while True:
                            buf = content.read(4096)
                            if not buf:
                                break
                            f.write(buf)
                for file in deleted:
                    path = osp.realpath(osp.join(tmpdir, file))
                    assert path.startswith(osp.realpath(tmpdir))
                    try:
                        os.unlink(path)
                    except OSError as exc:
                        if exc.errno == errno.ENOENT:
                            raise QueryError('%s is already deleted from the vcs' % file)
                        raise
                rev, cset = client.commit(message=description, addremove=True, user=author)
                client.push(rev=cset, force=True)
                return cset[:12]
        finally:
            shutil.rmtree(tmpdir)

    def pull(self):
        self._create_repo()
        if self.entity.source_url:
            with self.hgrepo() as hgrepo:
                try:
                    hgrepo.pull(source=self.entity.source_url)
                except CommandError as exc:
                    raise EnvironmentError('pull or clone from %s failed %s' %
                                           (self.entity.source_url, exc))

    def import_content(self, commitevery):
        cnx = self._cw
        try:
            self.pull()
        except EnvironmentError as exc:
            # repo at source_url may not be available, notify it but
            # go on since there might have new commits in local cache
            self.warning(str(exc))
        with self.hgrepo() as hgrepo:
            rset = cnx.execute('Any CSET WHERE R eid %(r)s, X from_repository R, '
                               'NOT CREV parent_revision X, X changeset CSET',
                               {'r': self.entity.eid})
            # we don't want no unicode
            knownheads = [str(row[0]) for row in rset]
            # detect stripped branches
            needsstrip = False
            while knownheads:
                # 100 at a time so we don't crash hg in a deep recursion
                csets = knownheads[:100]
                knownheads = knownheads[100:]
                try:
                    hgrepo.log(csets)
                except CommandError:
                    needsstrip = True
                    self.warning('%s: strip detected. One of %s is unknown in local cache',
                                 self.path, csets)
                    break

            if needsstrip:
                self.warning('%s: strip from %s.',
                             self.entity.dc_title(), self.path)
                revs = cnx.execute('Any R WHERE R is Revision, R from_repository REPO, '
                                   'REPO eid %(reid)s', {'reid': self.entity.eid})
                for reventity in revs.entities():
                    try:
                        hgrepo.log(reventity.changeset)
                    except CommandError:
                        reventity.cw_delete()

                cnx.commit()

            # update phases and visibility of already known revisions
            self.update_phases(cnx, hgrepo)
            self.update_hidden(cnx, hgrepo)
            # XXX we need to update obsolescence markers without new revision
            #     involved.
            # XXX the below would be a bit nicer, but makes mercurial explode due to too deep recursion
            #rset = cnx.execute(
            #        'Any CSET WHERE R eid %(r)s, X from_repository R, '
            #        'NOT CREV parent_revision X, X changeset CSET',
            #        {'r': self.entity.eid})
            #knownheads_cs = [str(row[0]) for row in rset if str(row[0]) in hgrepo]
            #revrange = ' and '.join(('not(::%s)' % i) for i in knownheads_cs) if knownheads_cs else None
            #missing = [int(rev[0]) for rev in hgrepo.log(revrange=revrange)]
            # XXX so we use this, which seems to need a lot of cpu instead :(
            rset = cnx.execute(
                    'Any CSET WHERE R eid %(r)s, X from_repository R, X changeset CSET',
                    {'r': self.entity.eid})
            knowncs = frozenset(str(row[0]) for row in rset)
            missing = sorted(int(rev[0]) for rev in hgrepo.log() if rev[1][:12] not in knowncs)
            missing.sort()
            while missing:
                for rev in missing[:commitevery]:
                    try:
                        self.import_revision(hgrepo, rev)
                    except Exception:
                        self.critical('error while importing revision %s of %s',
                                      rev, self.path, exc_info=True)
                        raise
                cnx.commit()
                del missing[:commitevery] # pop iterated value
            cnx.commit()

    def update_phases(self, cnx, hgrepo):
        """ update all phases (this potentially affects all revs since
        phase changes are not transactional) """
        repoeid = self.entity.eid
        for phase in (u'public', u'draft', u'secret'):
            revs = set(ctx.node[:12] for ctx in hgrepo.log(revrange='%s()' % phase))
            if revs:
                cnx.execute('SET R phase %%(phase)s WHERE R changeset IN (%s),'
                            'NOT R phase %%(phase)s, R from_repository REPO,'
                            'REPO eid %%(repo)s' %
                            ','.join(repr(rev) for rev in revs),
                            {'repo': repoeid, 'phase': phase})

    def update_hidden(self, cnx, hgrepo):
        """ maintain hidden status for all revs """
        repoeid = self.entity.eid
        oldhidden = set(num for num, in cnx.execute('Any CS WHERE R hidden True, '
                                                    'R from_repository H, '
                                                    'R changeset CS, H eid %(hg)s',
                                                    {'hg': repoeid}).rows)
        newhidden = set(rev.node[:12] for rev in hgrepo.log(revrange='hidden()'))
        hide = newhidden - oldhidden
        show = oldhidden - newhidden
        for revs, setto in ((hide, 'True'), (show, 'False')):
            if revs:
                cnx.execute('SET R hidden %s WHERE R from_repository H,'
                            'H eid %%(hg)s, R changeset IN (%s)' %
                            (setto, ','.join('"%s"' % cs for cs in revs)),
                            {'hg': self.entity.eid})

        # very partial handling of obsolescence relation added after a revision
        # issue #2731056
        # this tries to find explanation for changeset that becomes hidden
        # - does not work for changesets already hidden
        # - does not work for public changesets
        for new_hide in hide:
            ctx = hgrepo[new_hide]
            RQL = '''SET S obsoletes P
                     WHERE R eid %(r)s,
                     P from_repository R,
                     P changeset %(p)s,
                     S from_repository R,
                     S changeset %(s)s,
                     NOT S obsoletes P
            '''
            # unknown successors will create the relation at insertion time
            data = {'r': self.entity.eid,
                    'p': str(ctx)}
            successors_changesets = [cs.node[:12] for cs in hgrepo.log(revrange='successors(%s)' % ctx.node())]
            for succ in successors_changesets:
                data['s'] = str(succ)
                cnx.execute(RQL, data)

    def import_revision(self, hgrepo, i):
        self.info('importing revision %s from %s', i, self.entity.dc_title())
        cnx = self._cw
        execute = cnx.execute
        ctx = hgrepo[i]
        ctx_hex = ctx.node()[:12]
        if self.entity_for_cset(ctx_hex):
            self.warning('skip revision %s, seems already imported', ctx_hex)
            return
        revdata = {'date': ctx.date(),
                   'author': unicode(ctx.author(), self.entity.encoding),
                   'description': unicode(ctx.description(), self.entity.encoding),
                   'changeset': unicode(ctx_hex, self.entity.encoding),
                   'phase': unicode(ctx.phase()),
                   'branch': unicode(ctx.branch()),
                   'hidden': ctx.hidden(),
                   'vcstags': u','.join(ctx.tags()),
                   }
        # truncate author field if it's too long
        constraint = cnx.vreg.schema['Revision'].rdef('author').constraint_by_type('SizeConstraint')
        if constraint and constraint.max and len(revdata['author']) > constraint.max:
            revdata['author'] = revdata['author'][:constraint.max - 1] + u'…'
        #XXX workaround, parents may return None
        parent_changesets = hgrepo.parents(ctx_hex)
        parent_changesets = [n.node[:12] for n in parent_changesets] if parent_changesets else []
        precursors_changesets = [cs.node[:12] for cs in hgrepo.log(revrange='allprecursors(%s)' % i)]
        if not precursors_changesets:
            precursors = []
        else:
            precursors = execute(
                    'Any X WHERE X changeset XC, '
                    'X changeset IN (%s), X from_repository R, R eid %%(r)s'
                    % ','.join("'%s'" % cs for cs in precursors_changesets),
                    {'r': self.entity.eid})
        revdata['precursors'] = [p[0] for p in precursors]
        if not parent_changesets:
            parents = []
        else:
            parents = execute(
                    'Any X,XC WHERE X changeset XC, '
                    'X changeset IN (%s), X from_repository R, R eid %%(r)s'
                    % ','.join("'%s'"%cs for cs in parent_changesets),
                    {'r': self.entity.eid})
        revdata['parents'] = [r[0] for r in parents]
        reveid = bridge.import_revision(cnx, self.entity.eid, **revdata)



"""custom forms to submit new revision to the svn repository or to edit
some information about existing revisions.

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb.web import formwidgets as fw
from cubicweb.web.views import uicfg

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs
_rctl = uicfg.reledit_ctrl
_pvdc = uicfg.primaryview_display_ctrl

for attr in ('path', 'type', 'source_url', 'encoding'):
    _pvdc.tag_attribute(('Repository', attr), {'vid': 'attribute'})
_rctl.tag_attribute(('Repository', 'title'), {'reload': True})
_affk.tag_attribute(('Repository', 'source_url'), {'widget': fw.TextInput})
_affk.tag_attribute(('Repository', 'path'), {'widget': fw.TextInput})
_afs.tag_attribute(('Repository', 'local_cache'), 'main', 'hidden')




# -*- coding: utf-8 -*-
"""hooks for vcsfile content types

:organization: Logilab
:copyright: 2007-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import with_statement

__docformat__ = "restructuredtext en"

try:
    from cubicweb import _
except ImportError:
    _ = unicode

import os
import os.path as osp
import errno
import shutil

from logilab.common.textutils import unormalize
from cubicweb import QueryError, ValidationError
from cubicweb.server import hook, session, Service
from cubicweb.predicates import is_instance, score_entity, match_user_groups

from cubes.vcsfile import bridge

def repo_cache_dir(config):
    """Return the directory to use for repo cache

    This function create directory if necessary."""
    directory = config['local-repo-cache-root']
    if not osp.isabs(directory):
        directory = osp.join(config.appdatahome, directory)
    config.local_repo_cache_root = directory
    if not osp.exists(directory):
        try:
            os.makedirs(directory)
        except Exception:
            config.critical('could not find local repo cache directory; check '
                            'local-repo-cache-root option whose present value '
                            'is %r)', config['local-repo-cache-root'])
            raise
    return directory

def url_to_relative_path(url):
    scheme, url = url.split('://', 1)
    if scheme == 'file':
        return url.rstrip('/').rsplit('/', 1)[-1]
    else:
        try:
            return url.rstrip('/').rsplit('/', 1)[1]
        except IndexError:
            return '/'

def set_local_cache(vcsrepo):
    if vcsrepo.local_cache:
        return
    cachedir = osp.join(repo_cache_dir(vcsrepo._cw.vreg.config),
                        unicode(vcsrepo.eid))
    if not osp.exists(cachedir):
        try:
            os.makedirs(cachedir)
        except OSError:
            vcsrepo.exception('cant create repo cache directory %s', cachedir)
            return
    if vcsrepo.source_url:
        reponame = vcsrepo.source_url.rstrip('/').rsplit('/', 1)[-1]
    elif vcsrepo.title:
        reponame = unormalize(vcsrepo.title).replace(' ', '_')
    else:
        reponame = 'repository'
    vcsrepo.cw_edited['local_cache'] = osp.join(unicode(vcsrepo.eid), reponame)

def clone_to_local_cache(vcsrepo):
    handler = vcsrepo.cw_adapt_to('VCSRepo')
    url = vcsrepo.source_url
    try:
        handler.pull()
    except Exception:
        handler.exception('while trying to clone repo %s', url)
        msg = _('can not clone the repo from %(source)s, '
                'please check the source url')
        raise ValidationError(vcsrepo.eid, {'source_url': msg}, {'source': url})


def missing_relation_error(entity, rtype):
    # use __ since msgid recorded in cw, we don't want to translate it in
    # this cube
    __ = entity._cw._
    msg = __('at least one relation %(rtype)s is required on %(etype)s (%(eid)s)')
    errors = {'from_repository': msg % {'rtype': __(rtype),
                                        'etype': __(entity.__regid__),
                                        'eid': entity.eid}}
    return ValidationError(entity.eid, errors)

def missing_attribute_error(entity, attrs):
    msg = _('at least one attribute of %s must be set on a Repository')
    errors = {}
    for attr in attrs:
        errors[attr] = msg % ', '.join(attrs)
    return ValidationError(entity.eid, errors)


# initialization hooks #########################################################


class VCSFileInitHook(hook.Hook):
    """install attribute map on the system source sql generator and init
    configuration
    """
    __regid__ = 'vcsfile.hook.init'
    events = ('server_startup', 'server_maintenance')

    def __call__(self):
        cacheroot = repo_cache_dir(self.repo.config)
        if not os.access(cacheroot, os.W_OK):
            raise ValueError('directory %r is not writable (check "local-repo-'
                             'cache-root" option)' % cacheroot)
        hgrcpath = self.repo.config['hgrc-path']
        if hgrcpath is not None:
            os.environ['HGRCPATH'] = hgrcpath



class VCSFileStartHook(hook.Hook):
    """register task to regularly pull/import vcs repositories content"""
    __regid__ = 'vcsfile.hook.start'
    events = ('server_startup',)

    def __call__(self):
        # closure to avoid error on code reloading
        interval = self.repo.config.get('check-revision-interval')
        if interval < 0:
            return
        repo = self.repo
        def refresh_vcsrepos():
            from cubes.vcsfile.hooks import _refresh_repos as refresh
            return refresh(repo)
        self.repo.looping_task(interval, refresh_vcsrepos)

# internals to be able to create new vcs repo revision using rql queries #######

### Other hooks
###############################################################################

class SetRepoCacheBeforeAddHook(hook.Hook):
    """clone just after repo creation

    mostly helps the tests
    """
    __regid__ = 'clone_repo_after_add'
    __select__ = (hook.Hook.__select__ & is_instance('Repository'))
    events = ('before_add_entity',)
    category = 'vcsfile.cache'
    order = 2

    def __call__(self):
        repo = self.entity
        if repo.type == 'mercurial':
            set_local_cache(repo)


class UpdateRepositoryHook(hook.Hook):
    """add repository eid to vcs bridge cache

    - Check that the type attribute is immutable
    - refresh the cache on url changes

    XXX stuff claiming to be cache but actually necessary are EVIL"""
    __regid__ = 'vcsfile.update_repository_hook'
    __select__ = hook.Hook.__select__ & is_instance('Repository')
    events = ('before_update_entity', )
    # XXX no category, contains integrity, mandatory and vcsfile.cache behaviour

    def __call__(self):
        vcsrepo = self.entity
        if 'type' in vcsrepo.cw_edited:
            msg = self._cw._('updating type attribute of a repository isn\'t '
                             'supported. Delete it and add a new one.')
            raise ValidationError(vcsrepo.eid, {'type': msg})
        if vcsrepo.type == 'mercurial' and not vcsrepo.local_cache:
            # XXX rmdir on rollback
            set_local_cache(vcsrepo)
            clone_to_local_cache(vcsrepo)


class DeleteDirsOp(hook.DataOperationMixIn, hook.Operation):

    def postcommit_event(self):
        for cachedir in self.get_data():
            try:
                try:
                    shutil.rmtree(cachedir)
                    self.info('deleted repository cache at %s', cachedir)
                except OSError, exc:
                    if (exc.errno != errno.ENOENT
                        or getattr(exc, 'filename', None) != cachedir):
                        raise
            except Exception:
                self.exception('cannot delete repository cache at %s', cachedir)

class DeleteRepositoryHook(hook.Hook):
    __regid__ = 'vcsfile.add_update_repository_hook'
    __select__ = (hook.Hook.__select__ &
                  is_instance('Repository') &
                  score_entity(lambda x: x.local_cache))
    events = ('before_delete_entity',)
    # XXX mandatory hook?

    def __call__(self):
        if self.entity.local_cache:
            cachedir = osp.dirname(osp.join(
                self._cw.vreg.config.local_repo_cache_root,
                self.entity.local_cache)).encode('ascii')
            DeleteDirsOp.get_instance(self._cw).add_data(cachedir)


# safety belts #################################################################

def _check_in_transaction(vf_or_rev):
    """check that a newly added versioned file or revision entity is done in
    a vcs repository transaction.
    """
    try:
        vcsrepo = vf_or_rev.from_repository[0]
    except IndexError:
        raise missing_relation_error(vf_or_rev, 'from_repository')
    try:
        transactions = vcsrepo._cw.transaction_data['vctransactions']
        transaction = transactions[vcsrepo.eid]
    except KeyError:
        raise QueryError('no transaction in progress for repository %s'
                         % vcsrepo.eid)
    return transaction


class AddUpdateRepositoryHook(hook.Hook):
    __regid__ = 'vcsfile.add_update_repository_hook'
    __select__ = hook.Hook.__select__ & is_instance('Repository')
    events = ('after_add_entity', 'after_update_entity')
    category = 'integrity'
    required_attrs = ('source_url',)
    order = 1

    def __call__(self):
        entity = self.entity
        if entity.source_url:
            try:
                url_to_relative_path(entity.source_url)
            except Exception:
                msg = self._cw._('url must be of the form protocol://path/to/stuff')
                raise ValidationError(entity.eid, {'source_url': msg})
        # ensure local cache exists
        entity.cw_adapt_to('VCSRepo')

def _refresh_repos(repo, eids=None, bridge=bridge, raise_on_error=False):
    '''
    Refresh vcs repos designated by the `eids` parameter, if refreshable:

      * None for all vcs repos
      * a list of integers for specific vcs repos
    '''
    with repo.internal_cnx() as cnx:
        config = cnx.vreg.config
        rql = ('Any R, RT, RE, RLC, RSU WHERE R is Repository, '
               'NOT (R local_cache NULL),'
               'R type RT, R encoding RE, R local_cache RLC, R source_url RSU')
        if not config['cache-external-repositories']:
            rql += ', R cw_source S, S name "system"'
        kwargs = None
        if eids is not None:
            if isinstance(eids, int):
                rql += ', R eid %(x)s'
                kwargs = {'x': eids}
            else:
                rql += ', R eid IN (%s)' % ','.join(str(eid) for eid in eids)
        reposrset = cnx.execute(rql, kwargs)
        for vcsrepo in reposrset.entities():
            # watch for repository being shut-down, our thread may:
            # * not receive the KeyboardInterrupt or shutdown signal
            # * loop a while without accessing the session (actually
            #   until some repository is actually not up-to-date, while
            #   this access to the session would abort the task)
            try:
                repohdlr = vcsrepo.cw_adapt_to('VCSRepo')
            except bridge.VCSException, ex:
                if raise_on_error:
                    raise
                cnx.error(str(ex))
                continue
            except Exception:
                if raise_on_error:
                    raise
                cnx.exception('error while retrieving handler for %s', vcsrepo.eid)
                continue
            if vcsrepo.local_cache is not None:
                try:
                    repohdlr.pull()
                except bridge.VCSException, ex:
                    if raise_on_error:
                        raise
                    repohdlr.error(str(ex))
                except Exception:
                    if raise_on_error:
                        raise
                    repohdlr.exception(
                        'error while updating local cache of repository %s',
                        vcsrepo.eid)
            if (config['repository-import']
                and vcsrepo.cw_metainformation()['source']['uri'] == 'system'):
                bridge.import_vcsrepo_content(
                    cnx, repohdlr, raise_on_error=raise_on_error,
                    commitevery=config.get('check-revision-commit-every', 1))


class RefreshRepoService(Service):
    __regid__ = 'refresh_repository'
    __select__ = Service.__select__ & match_user_groups('managers')

    def call(self, eids=None, raise_on_error=False):
        _refresh_repos(self._cw.repo, eids, raise_on_error=raise_on_error)


class VCSRepoNotifHook(hook.Hook):
    events = ('server_startup',)
    __regid__ = 'vcsfile_repo_notif_startup'
    order = 10 # make sure it is loaded "late" (in fact, it must be
               # loaded after ZMQStartHook but we lack a dependency
               # system here, see http://www.cubicweb.org/ticket/2900987)
    def __call__(self):
        def callback(msg):
            self.debug('received notification: %s', ' '.join(msg))
            msg = msg[-1]
            with self.repo.internal_cnx() as cnx:
                for suffix in ('', '/'):
                    rset = cnx.execute('Any R WHERE R source_url U, '
                                            'R source_url LIKE %(u)s',
                                            {'u': '%/'+msg+suffix})
                    if rset:
                        break
                if rset:
                    assert len(rset) == 1, 'found multiple repos for "%s"'%msg
                    eid = rset[0][0]
                    self.debug('Calling refresh service for repo %s', eid)
                    cnx.call_service('refresh_repository', eids=[eid])
                else:
                    self.debug('No repo for %s', msg)

        self.repo.app_instances_bus.add_subscription('vcsfile-repo-notif', callback)



# -*- coding: utf-8 -*-
"""makes the bridge between repository handlers and related cubicweb entities

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import with_statement

__docformat__ = "restructuredtext en"

from yams import ValidationError

from cubes.vcsfile import queries


class VCSException(Exception):
    def __init__(self, repoeid, attr, msgid, msgargs=()):
        self.repoeid = repoeid
        self.attr = attr
        self.msgid = msgid
        self.msgargs = msgargs

    def __str__(self):
        return '%s.%s: %s' % (self.repoeid, self.attr,
                              self.msgid % self.msgargs)

    def to_validation_error(self, translate):
        msg = translate(self.msgid)
        if self.msgargs:
            msg %= self.msgargs
        return ValidationError(self.repoeid, {self.attr: msg})


def import_content(cnx, commitevery=1000, raise_on_error=False):
    """synchronize content of known vcs repositories
    """
    for vcsrepo in cnx.execute(
        'Any X,T,U,L,E WHERE X is Repository, '
        'X type T, X source_url U, X local_cache L, '
        'X encoding E').entities():
        try:
            repohdlr = vcsrepo.cw_adapt_to('VCSRepo')
        except VCSException, ex:
            if raise_on_error:
                raise
            cnx.error(str(ex))
            continue
        import_vcsrepo_content(cnx, repohdlr,
                               commitevery=commitevery,
                               raise_on_error=raise_on_error)


def import_vcsrepo_content(cnx, repohdlr, commitevery=10,
                           raise_on_error=False):
    """synchronize content of given vcs repository"""
    vcsrepo = repohdlr.entity
    try:
        repohdlr.import_content(commitevery)
        # give other threads a chance to run: commit will release the cnxset, get
        # it back afterwards
        cnx.commit()
    except Exception:
        if raise_on_error:
            raise
        try:
            title = vcsrepo.dc_title()
        except Exception:
            title = str(vcsrepo)
        repohdlr.exception(
            'error while importing content for vcs repo %s', title)
        cnx.rollback()


def import_revision(cnx, repoeid, date, **kwargs):
    """import a new revision from a repository"""
    args = {'date': date}
    args['repoeid'] = repoeid
    for attr in ('author', 'description', 'changeset', 'branch', 'phase', 'hidden', 'vcstags'):
        args[attr] = kwargs.get(attr)
    prevs = kwargs['parents']
    if prevs:
        args['parent'] = prevs[0]
    precs = kwargs['precursors']
    if precs:
        args['precursor'] = precs[0]
    reveid = cnx.execute(queries.new_revision_rql(prevs, precs, True), args)[0][0]
    # complete other parents & precursors relations
    if len(prevs) > 1:
        for preveid in prevs[1:]:
            cnx.execute('SET R parent_revision PR WHERE R eid %(r)s, PR eid %(pr)s',
                        {'r': reveid, 'pr': preveid})
    if len(precs) > 1:
        for preceid in precs[1:]:
            cnx.execute('SET R obsoletes PREC WHERE R eid %(r)s, PREC eid %(pr)s',
                        {'r': reveid, 'pr': preceid})
    return reveid



"template automatic tests"
import os
import os.path as osp
from cubicweb.devtools import testlib

os.environ['HGRCPATH'] = os.devnull

class AutomaticWebTest(testlib.AutomaticWebTest):
    no_auto_populate = ('Repository', 'Revision')
    ignored_relations = set(('parent_revision',
                             'from_repository'))

    def to_test_etypes(self):
        return set(('Repository',
                    'Revision'))

    def custom_populate(self, how_many, cnx):
        cnx.create_entity('Repository', type=u'mercurial',
                          source_url=u'file://'+unicode(osp.join(self.datadir, 'testrepohg')),
                          encoding=u'latin1')

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()
